// Copyright (c) 2022 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#pragma once
#include "SaveFile.hpp"
#include "util/string.hpp"

#define SAVE_KEY_TIMESTAMP "meta.timestamp"
#define SAVE_KEY_SLOT "meta.slot"
#define SAVE_MANAGER_BUFFER_LENGTH 128
#define SAVE_MANAGER_FILENAME_LENGTH 512

namespace Dawn {
  class DawnGame;

  enum SaveLoadResult {
    SAVE_LOAD_SUCCESS,
    SAVE_LOAD_RESULT_FILE_NOT_PRESENT,
    SAVE_LOAD_RESULT_TOO_SMALL,
    SAVE_LOAD_RESULT_CORRUPTED_DE_SAVE,
    SAVE_LOAD_RESULT_CORRUPTED_VERSION,
    SAVE_LOAD_RESULT_CORRUPTED_TIMESTAMP,
    SAVE_LOAD_RESULT_CORRUPTED_DATA_KEY,
    SAVE_LOAD_RESULT_CORRUPTED_DATA_VALUE,
    SAVE_LOAD_RESULT_MISMATCH_TIMESTAMP,
    SAVE_LOAD_RESULT_ERROR
  };

  /**
   * Reads a save file and then validates all of the values. If a value is 
   * not validated then it will NOT be retained.
   * 
   * @param file Save file to read from.
   * @return Whether or not the save file is corrupt, false for not corrupt.
   */
  bool_t saveValidateFile(struct SaveFile raw);

  class SaveManager {
    protected:
      int16_t currentSaveSlot = -1;
      struct SaveFile currentSave;

    public:
      DawnGame *game;

      /**
       * Create a new Save Manager. Used by the game instance to decide when/how
       * to perform saving.
       * 
       * @param game Game that this save manager belongs to.
       */
      SaveManager(DawnGame *game);

      /**
       * Immediately save the game to the given file. This will invoke the 
       * game's custom writer to write the output data to the save file.
       */
      void saveFile();

      /**
       * Loads the current slotted save file. Invokes the internal managers read
       * function.
       */
      enum SaveLoadResult loadFile();

      /**
       * Deletes the given save file slot. We use slots here because we need to
       * be able to delete a file without reading it (in case of corruption).
       * 
       * @param slot Slot to delete.
       */
      void deleteFile(int16_t slot);

      /**
       * Determines which save slot to use.
       * 
       * @param slot Save slot to use.
       */
      void useSlot(int16_t slot);

      /**
       * Returns a list of used save slots, does not confirm if they are corrupt
       * or not, just whether they are in existance or not.
       * 
       * @return List of used save slots.
       */
      std::vector<int16_t> getUsedSlots();
  };
}