// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "TextureRenderTarget.hpp"

using namespace Dawn;

TextureRenderTarget::TextureRenderTarget(float_t width, float_t height) {
  this->setSize(width, height);
}

Texture * TextureRenderTarget::getTexture() {
  return &this->texture;
}

void TextureRenderTarget::setSize(float_t width, float_t height) {
  assertTrue(width > 0);
  assertTrue(height > 0);

  if(width == this->getWidth() && height == this->getHeight()) return;

  // Delete old buffers.
  if(this->rboId != -1) glDeleteRenderbuffers(1, &this->rboId);
  if(this->fboId != -1) glDeleteFramebuffers(1, &this->fboId);

  // Resize texture
  this->texture.setSize((int32_t)width, (int32_t)height, TEXTURE_FORMAT_RGBA);
  this->eventRenderTargetResized.invoke(this, width, height);

  // Create Frame Buffer
  glGenFramebuffers(1, &this->fboId);
  glBindFramebuffer(GL_FRAMEBUFFER, this->fboId);
  glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0,
    GL_TEXTURE_2D, this->texture.id, 0
  );

  // Create Render Buffer
  glGenRenderbuffers(1, &this->rboId);
  glBindRenderbuffer(GL_RENDERBUFFER, this->rboId); 
  glRenderbufferStorage(
    GL_RENDERBUFFER, GL_DEPTH24_STENCIL8,
    this->texture.width, this->texture.height
  );
  glBindRenderbuffer(GL_RENDERBUFFER, 0);
  glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_STENCIL_ATTACHMENT,
    GL_RENDERBUFFER, this->rboId
  );

  // Validate things went correct.
  if(glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE) {
    assertUnreachable();
  }
}

float_t TextureRenderTarget::getWidth() {
  return (float_t)this->texture.getWidth();
}

float_t TextureRenderTarget::getHeight() {
  return (float_t)this->texture.getHeight();
}

void TextureRenderTarget::setClearColor(struct Color color) {
  this->clearColor = color;
}

void TextureRenderTarget::clear(flag8_t clearFlags) {
  glClearColor(
    clearColor.r,
    clearColor.g,
    clearColor.b,
    clearColor.a
  );
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

void TextureRenderTarget::bind() {
  glBindFramebuffer(GL_FRAMEBUFFER, this->fboId);
  glViewport(0, 0, this->texture.getWidth(), this->texture.getHeight());
}

TextureRenderTarget::~TextureRenderTarget() {
  if(this->rboId != -1) glDeleteRenderbuffers(1, &this->rboId);
  if(this->fboId != -1) glDeleteFramebuffers(1, &this->fboId);
}